﻿using System;
using System.Globalization;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;
using Efir.Client.Communication;

namespace PartnerDataClient
{
	/// <summary>
	/// Реализация интерфейса IConnectionInfo для подключения к серверам Efir
	/// </summary>
	public class ConnectionInfo : IConnectionInfo
	{
		public ConnectionInfo()
		{
			UserInfo = new UserInfo();
			OndemandProxyInfo = ProxySettings.Empty;
			OnlineProxyInfo = ProxySettings.Empty;
		}

		/// <summary>
		/// Конструктор параметров подключения к online-серверу Efir
		/// </summary>
		/// <param name="srv">Адрес online-сервера</param>
		/// <param name="login">Имя пользователя Efir</param>
		/// <param name="pwd">Пароль</param>
		/// <param name="alias"></param>
		public ConnectionInfo(string srv, string login, string pwd, string alias)
			: this()
		{
			UserInfo = new UserInfo { Name = login, Psw = pwd };
			OnlineHostInfo = new TCPInfo
								 {
									 Host = srv.Split(':')[0],
									 Port = int.Parse(srv.Split(':')[1]),
									 Name = alias,
									 Proto = TCPInfoProto.http
								 };
		}

		public IUserInfo UserInfo { get; set; }
		public ITCPInfo ConfigurationHostInfo { get; set; }
		public IProxySettings OndemandProxyInfo { get; set; }
		public IProxySettings OnlineProxyInfo { get; set; }
		public ITCPInfo OnlineHostInfo { get; set; }
		public ITCPInfo OnDemandHostInfo { get; set; }

		public object Clone()
		{
			var rv = new ConnectionInfo
						 {
							 UserInfo = UserInfo.Clone() as IUserInfo,
							 ConfigurationHostInfo = ConfigurationHostInfo.Clone() as ITCPInfo,
							 OndemandProxyInfo = OndemandProxyInfo.Clone() as IProxySettings,
							 OnlineProxyInfo = OnlineProxyInfo.Clone() as IProxySettings
						 };
			if (OnlineHostInfo != null)
				rv.OnlineHostInfo = OnlineHostInfo.Clone() as ITCPInfo;
			if (OnDemandHostInfo != null)
				rv.OnDemandHostInfo = OnDemandHostInfo.Clone() as ITCPInfo;
			return rv;
		}

		public XmlSchema GetSchema()
		{
			return null;
		}

		public void ReadXml(XmlReader reader)
		{
			var subreader = reader.ReadSubtree();
			while (subreader.Read())
			{
				if (!subreader.IsStartElement() || subreader.Depth != 1) continue;
				switch (subreader.LocalName)
				{
					case "User":
						{
							UserInfo = new UserInfo();
							UserInfo.ReadXml(reader);
						}
						break;
					case "ConfigurationHost":
						{
							ConfigurationHostInfo = new TCPInfo();
							ConfigurationHostInfo.ReadXml(reader);
						}
						break;
					case "OnlineHost":
						{
							OnlineHostInfo = new TCPInfo();
							OnlineHostInfo.ReadXml(reader);
						}
						break;
					case "OnDemandHost":
						{
							OnDemandHostInfo = new TCPInfo();
							OnDemandHostInfo.ReadXml(reader);
						}
						break;
					case "OnDemandProxy":
						{
							OndemandProxyInfo = new ProxySettings();
							OndemandProxyInfo.ReadXml(reader);
						}
						break;
					case "OnlineProxy":
						{
							OnlineProxyInfo = new ProxySettings();
							OnlineProxyInfo.ReadXml(reader);
						}
						break;
				}
			}
			reader.ReadEndElement();
			if (ConfigurationHostInfo != null && ConfigurationHostInfo.Proto != TCPInfoProto.custom)
			{
				OnlineHostInfo = null;
				OnDemandHostInfo = null;
			}
		}

		public void WriteXml(XmlWriter writer)
		{
			var ns = new XmlSerializerNamespaces();
			ns.Add("", "");

			if (UserInfo != null)
			{
				writer.WriteStartElement("User");
				UserInfo.WriteXml(writer);
				writer.WriteEndElement();
			}

			if (ConfigurationHostInfo != null)
			{
				writer.WriteStartElement("ConfigurationHost");
				ConfigurationHostInfo.WriteXml(writer);
				writer.WriteEndElement();
			}

			if (OndemandProxyInfo != null)
			{
				writer.WriteStartElement("OnDemandProxy");
				OndemandProxyInfo.WriteXml(writer);
				writer.WriteEndElement();
			}

			if (OnlineProxyInfo != null)
			{
				writer.WriteStartElement("OnlineProxy");
				OnlineProxyInfo.WriteXml(writer);
				writer.WriteEndElement();
			}

			if (ConfigurationHostInfo != null && ConfigurationHostInfo.Proto == TCPInfoProto.custom)
			{
				writer.WriteStartElement("OnlineHost");
				OnlineHostInfo.WriteXml(writer);
				writer.WriteEndElement();

				writer.WriteStartElement("OnDemandHost");
				OnDemandHostInfo.WriteXml(writer);
				writer.WriteEndElement();
			}
		}
	}

	[XmlRoot("UserInfo")]
	public class UserInfo : IUserInfo
	{
		[XmlElement("user")]
		public string Name { get; set; }

		[XmlElement("psw")]
		public string Psw { get; set; }

		public bool IsValid()
		{
			return (!String.IsNullOrEmpty(Name) && (Psw != null));
		}

		public virtual object Clone()
		{
			return new UserInfo { Name = Name, Psw = Psw };
		}

		public XmlSchema GetSchema()
		{
			return null;
		}

		public void ReadXml(XmlReader reader)
		{
			Name = reader["user"];
			Psw = reader["psw"];
		}

		public void WriteXml(XmlWriter writer)
		{
			writer.WriteAttributeString("user", Name);
			writer.WriteAttributeString("psw", Psw);
		}
	}

	[XmlRoot("Proxy")]
	public class ProxySettings : IProxySettings
	{
		public static IProxySettings Empty { get; private set; }

		static ProxySettings()
		{
			Empty = new ProxySettings { GateType = TCPGateType.gtNone };
		}

		[XmlElement("GateType")]
		public TCPGateType GateType { get; set; }

		[XmlElement("Host")]
		public string Host { get; set; }

		[XmlElement("Port")]
		public int Port { get; set; }

		[XmlElement("Username")]
		public string Username { get; set; }

		[XmlElement("Password")]
		public string Password { get; set; }

		[XmlIgnore]
		public bool UseAuthorization { get { return !string.IsNullOrEmpty(Username); } }

		public object Clone()
		{
			var rv = new ProxySettings
				         {
				GateType = GateType,
				Host = Host,
				Password = Password,
				Port = Port,
				Username = Username
			};
			return rv;
		}

		public XmlSchema GetSchema()
		{
			return null;
		}

		public void ReadXml(XmlReader reader)
		{
			GateType = (TCPGateType)Enum.Parse(typeof(TCPGateType), (reader["gate"] ?? TCPGateType.gtNone.ToString()));
			Host = reader["host"];
			Port = int.Parse(reader["port"] ?? "0");
			Username = reader["user"];
			Password = reader["password"];
		}

		public void WriteXml(XmlWriter writer)
		{
			if (GateType != TCPGateType.gtNone) writer.WriteAttributeString("gate", GateType.ToString());
			if (!string.IsNullOrEmpty(Host)) writer.WriteAttributeString("host", Host);
			if (Port != 0) writer.WriteAttributeString("port", Port.ToString(CultureInfo.InvariantCulture));
			if (!string.IsNullOrEmpty(Username)) writer.WriteAttributeString("user", Username);
			if (!string.IsNullOrEmpty(Password)) writer.WriteAttributeString("password", Password);
		}

		public override string ToString()
		{
			if (GateType == TCPGateType.gtNone)
			{
				return "GateType=" + GateType.ToString();
			}
			return string.Format("GateType={0}, Host={1}, Port={2}, User={3}", GateType, Host, Port, Username);
		}
	}

	[XmlRoot("TCPInfo")]
	public class TCPInfo : ITCPInfo
	{
		[XmlAttribute("name")]
		public string Name { get; set; }

		[XmlElement("host")]
		public string Host { get; set; }

		[XmlElement("port")]
		public int Port { get; set; }

		[XmlElement("FirewallTimeout")]
		public int FirewallTimeout { get; set; }

		[XmlElement("proto")]
		public TCPInfoProto Proto { get; set; }

		public virtual object Clone()
		{
			var rv = new TCPInfo
				         {
				Name = Name,
				Host = Host,
				Port = Port,
				FirewallTimeout = FirewallTimeout,
				Proto = Proto,
			};
			return rv;
		}

		public XmlSchema GetSchema()
		{
			return null;
		}

		public void ReadXml(XmlReader reader)
		{
			var subReader = reader.ReadSubtree();
			var thisClone = new TCPInfo();

			while (subReader.Read())
			{
				subReader.MoveToContent();
				if (!subReader.IsStartElement() || subReader.Depth != 1) continue;
				switch (subReader.LocalName)
				{
					case "name":
						thisClone.Name = subReader.ReadString();
						break;
					case "host":
						thisClone.Host = subReader.ReadString();
						break;
					case "port":
						thisClone.Port = int.Parse(subReader.ReadString());
						break;
					case "FirewallTimeout":
						thisClone.FirewallTimeout = int.Parse(subReader.ReadString());
						break;
					case "proto":
						thisClone.Proto = (TCPInfoProto)Enum.Parse(typeof(TCPInfoProto), subReader.ReadString(), true);
						break;
				}
			}

			Name = thisClone.Name;
			Host = thisClone.Host;
			Port = thisClone.Port;
			FirewallTimeout = thisClone.FirewallTimeout;
			Proto = thisClone.Proto;
		}

		public void WriteXml(XmlWriter writer)
		{
			writer.WriteElementString("name", Name);
			writer.WriteElementString("host", Host);
			if (Port != 0) writer.WriteElementString("port", Port.ToString(CultureInfo.InvariantCulture));
			writer.WriteElementString("proto", Proto.ToString());
			if (FirewallTimeout != 0) writer.WriteElementString("FirewallTimeout", FirewallTimeout.ToString(CultureInfo.InvariantCulture));
		}
	}
}
