using System;
using System.Runtime.InteropServices;
using System.Text;
using Microsoft.Win32.SafeHandles;

namespace PartnerDataClient
{
	/// <summary>
	///  WinAPI-    WindowsCredentialManager
	/// </summary>
	public class CredentialsApi
	{
		[DllImport("Advapi32.dll", EntryPoint = "CredReadW", CharSet = CharSet.Unicode, SetLastError = true)]
		private static extern bool CredRead(string target, CRED_TYPE type, int reservedFlag, out IntPtr credentialPtr);

		[DllImport("Advapi32.dll", EntryPoint = "CredWriteW", CharSet = CharSet.Unicode, SetLastError = true)]
		private static extern bool CredWrite([In] ref NativeCredential userCredential, [In] UInt32 flags);

		[DllImport("Advapi32.dll", EntryPoint = "CredFree", SetLastError = true)]
		private static extern bool CredFree([In] IntPtr cred);

		/// <summary>
		///   
		/// </summary>
		public enum CRED_TYPE : uint
		{
			GENERIC = 1,
			DOMAIN_PASSWORD = 2,
			DOMAIN_CERTIFICATE = 3,
			DOMAIN_VISIBLE_PASSWORD = 4
		}

		/// <summary>
		///    
		/// </summary>
		public enum CRED_PERSIST : uint
		{
			SESSION = 1,
			LOCAL_MACHINE = 2,
			ENTERPRISE = 3
		}

		/// <summary>
		///  ,   
		/// </summary>
		[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
		private struct NativeCredential
		{
			public UInt32 Flags;
			public UInt32 Type;
			public IntPtr TargetName;
			public IntPtr Comment;
			public System.Runtime.InteropServices.ComTypes.FILETIME LastWritten;
			public UInt32 CredentialBlobSize;
			public IntPtr CredentialBlob;
			public UInt32 Persist;
			public UInt32 AttributeCount;
			public IntPtr Attributes;
			public IntPtr TargetAlias;
			public IntPtr UserName;

			/// <summary>
			/// This method derives a NativeCredential instance from a given Credential instance.
			/// </summary>
			/// <param name="cred">The managed Credential counterpart containing data to be stored.</param>
			/// <returns>A NativeCredential instance that is derived from the given Credential instance.</returns>
			internal static NativeCredential GetNativeCredential(Credential cred)
			{
				NativeCredential ncred = new NativeCredential
					                         {
						                         AttributeCount = 0,
						                         Attributes = IntPtr.Zero,
						                         Comment = IntPtr.Zero,
						                         TargetAlias = IntPtr.Zero,
						                         Type = (UInt32) cred.Type,
						                         Persist = (UInt32) cred.Persist,
						                         CredentialBlobSize = cred.CredentialBlobSize,
						                         TargetName = Marshal.StringToCoTaskMemUni(cred.TargetName),
						                         CredentialBlob = Marshal.StringToCoTaskMemUni(cred.CredentialBlob),
						                         UserName = Marshal.StringToCoTaskMemUni(cred.UserName)
					                         };
				return ncred;
			}
		}

		[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
		private struct Credential
		{
			public UInt32 Flags;
			public CRED_TYPE Type;
			public string TargetName;
			public string Comment;
			public System.Runtime.InteropServices.ComTypes.FILETIME LastWritten;
			public UInt32 CredentialBlobSize;
			public string CredentialBlob;
			public CRED_PERSIST Persist;
			public UInt32 AttributeCount;
			public IntPtr Attributes;
			public string TargetAlias;
			public string UserName;
		}

		#region Public methods
		/// <summary>
		///  generic      Credential Manager
		/// </summary>
		/// <param name="key">  </param>
		/// <param name="userName"> </param>
		/// <param name="password"></param>
		public static void WriteCred(string key, string userName, string password)
		{
			byte[] byteArray = Encoding.Unicode.GetBytes(password);
			if (byteArray.Length > 512)
				throw new ArgumentOutOfRangeException("password");

			Credential cred = new Credential
				                  {
					                  TargetName = key,
					                  UserName = userName,
					                  CredentialBlob = password,
					                  CredentialBlobSize = (UInt32) byteArray.Length,
					                  AttributeCount = 0,
					                  Attributes = IntPtr.Zero,
					                  Comment = null,
					                  TargetAlias = null,
					                  Type = CRED_TYPE.GENERIC,
					                  Persist = CRED_PERSIST.LOCAL_MACHINE
				                  };
			NativeCredential ncred = NativeCredential.GetNativeCredential(cred);
			bool written = CredWrite(ref ncred, 0);
			if (!written)
			{
				throw new System.ComponentModel.Win32Exception(Marshal.GetLastWin32Error());
			}
		}

		/// <summary>
		///   generic     
		/// </summary>
		/// <param name="key">  </param>
		/// <param name="userName"> </param>
		/// <param name="password"></param>
		public static void ReadCred(string key, out string userName, out string password)
		{
			IntPtr nCredPtr;
			userName = null;
			password = null;
			bool read = CredRead(key, CRED_TYPE.GENERIC, 0, out nCredPtr);
			if (!read)
			{
				throw new System.ComponentModel.Win32Exception(Marshal.GetLastWin32Error());
			}
			using (CriticalCredentialHandle critCred = new CriticalCredentialHandle(nCredPtr))
			{
				Credential cred = critCred.GetCredential();
				userName = cred.UserName;
				password = cred.CredentialBlob;
			}
		}

		#endregion

		#region Critical Handle Type definition
		/// <summary>
		///      
		/// </summary>
		sealed class CriticalCredentialHandle : CriticalHandleZeroOrMinusOneIsInvalid
		{
			internal CriticalCredentialHandle(IntPtr handle)
			{
				SetHandle(handle);
			}

			/// <summary>
			///       
			/// </summary>
			/// <returns></returns>
			internal Credential GetCredential()
			{
				if (!IsInvalid)
				{
					NativeCredential ncred = (NativeCredential)Marshal.PtrToStructure(handle,
						  typeof(NativeCredential));

					Credential cred = new Credential
						                  {
							                  CredentialBlobSize = ncred.CredentialBlobSize,
							                  CredentialBlob = Marshal.PtrToStringUni(ncred.CredentialBlob,
							                                                          (int) ncred.CredentialBlobSize / 2),
							                  UserName = Marshal.PtrToStringUni(ncred.UserName),
							                  TargetName = Marshal.PtrToStringUni(ncred.TargetName),
							                  TargetAlias = Marshal.PtrToStringUni(ncred.TargetAlias),
							                  Type = (CRED_TYPE) ncred.Type,
							                  Flags = ncred.Flags,
							                  Persist = (CRED_PERSIST) ncred.Persist
						                  };
					return cred;
				}
				throw new InvalidOperationException("Invalid CriticalHandle!");
			}

			/// <summary>
			///   ,     
			/// </summary>
			/// <returns>True -  . False -    .</returns>
			override protected bool ReleaseHandle()
			{
				if (!IsInvalid)
				{
					CredFree(handle);
					//    
					SetHandleAsInvalid();
					return true;
				}
				return false;
			}
		}
		#endregion
	}
}