﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Configuration;
using System.Diagnostics;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Efir.Client.Communication;
using Interfax.Efir.MarketModel;
using PartnerDataClient.Properties;

namespace PartnerDataClient.Pages
{
	/// <summary>
	/// Пример bulk-подписки на онлайн-данные
	/// </summary>
	public partial class BulkOnlinePage : Form
	{
		private const string ClentVersion = "1.0.0.0";	//"7.9.4.50576";

		// Названия секций в ответах сервера
		const string SectionResult = "RESULT";
		const string SectionStaticFields = "HIDE_ONLINE";	// Массив флагов, отмещающих статические поля.
		const string SectionParams = "PARAM";		// Параметры запроса
		const string SectionData = "DATA";			// Собственно данные по инструмента
		const string SectionGlobalData = "GDATA";	// Пристутствует только для стаканов: в ней лежит id_iss

		private ServerConnectionAdmin _connAdmin;	// Менеджер коннектов к серверу
		private IServerSession _requestSession;		// Сессия
		private Status _status = Status.Offline;	// Текущее состояние сессии
		private bool _isRunning;					// Флаг активности
		private CancellationTokenSource _cancellation = new CancellationTokenSource();
		private Request _currentRequest;			// Параметры текущего выполняемого запроса
		private string[] _columns;					// Список колонок, полученный с сервера

		/// <summary>
		/// Состояние сессии
		/// </summary>
		private enum Status
		{
			Offline,
			Connecting,
			LoggedIn,
			Connected,
			Disconnected,
			Error
		}

		/// <summary>
		/// Параметры запроса, введённые пользователем
		/// </summary>
		private class Request
		{
			/// <summary>
			/// Торговая площадка
			/// </summary>
			public MarketFeedGroup Market { get; set; }
			/// <summary>
			/// (Необязательный) Список колонок
			/// </summary>
			public string Columns { get; set; }
			/// <summary>
			/// (Необязательный) Список инструментов
			/// </summary>
			public string Instruments { get; set; }
		}

		public BulkOnlinePage()
		{
			InitializeComponent();
		}

		protected override void OnLoad(EventArgs e)
		{
			base.OnLoad(e);
			SetStatus(Status.Offline);
			// Заполняем выпадающий список значениями MarketFeedGroup
			List<string> groups = new List<string>(Enum.GetNames(typeof(MarketFeedGroup)));
			groups.Sort();
			foreach (var name in groups)
			{
				cbMarketFeedGroup.Items.Add(name);
			}
			cbMarketFeedGroup.SelectedItem = "MicexShares";
		}

		protected override void OnClosing(CancelEventArgs e)
		{
			if (_isRunning)
			{
				// Если выполнялся запрос - посылаем сигнал и дожидаемся его остановки
				Debug.WriteLine("Stopping BulkOnlinePage");
				_cancellation.Cancel();
				while (_isRunning)
				{
					Application.DoEvents();
				}
				Debug.WriteLine("BulkOnlinePage is stopped");
			}
			base.OnClosing(e);
		}

		/// <summary>
		/// Подключение к online-серверу
		/// </summary>
		private void Connect()
		{
			try
			{
				SetStatus(Status.Connecting);
				_connAdmin = ServerConnectionAdmin.Create(ClentVersion, AsyncOperationManager.SynchronizationContext);
				_connAdmin.OnCustomCommand += ServerAdmin_OnCustomCommand;
				_connAdmin.OnErrors += ServerAdmin_OnErrors;
				_connAdmin.OnStateChanged += ServerAdmin_OnStateChanged;

				string onlineServerAddress = ConfigurationManager.AppSettings["OnlineServer"];
				ConnectionInfo connectionInfo = new ConnectionInfo(onlineServerAddress, Program.UserName, Program.Password, "NEWS");
				_connAdmin.ConnectFor(connectionInfo);
			}
			catch (Exception ex)
			{
				SetStatus(Status.Error);
				AppendText(ex.ToString());
			}
		}

		/// <summary>
		/// Отключение от сервера - закрываем сессию и коннект к серверу
		/// </summary>
		private void Disconnect()
		{
			if (_requestSession != null)
			{
				if (_requestSession.IsConnected)
				{
					_requestSession.Disconnect();
				}
				_requestSession.OnConnect -= Session_OnConnect;
				_requestSession.OnData -= Session_OnData;
				_requestSession.OnDataBulk -= Session_OnDataBulk;
				_requestSession.OnDisconnect -= Session_OnDisconnect;
				_requestSession.OnError -= Session_OnError;
				_requestSession.Dispose();
				_requestSession = null;
			}
			if (_connAdmin != null)
			{
				if (_connAdmin.Connected)
				{
					_connAdmin.Disconnect();
				}
				_connAdmin.OnCustomCommand -= ServerAdmin_OnCustomCommand;
				_connAdmin.OnErrors -= ServerAdmin_OnErrors;
				_connAdmin.OnStateChanged -= ServerAdmin_OnStateChanged;
				_connAdmin.Dispose();
				_connAdmin = null;
			}
			SetStatus(Status.Offline);
		}

		/// <summary>
		/// Обработчик системных команд от сервера.
		/// Чаще всего это бывает команда SYS_TIME, параметром которой является серверное время в UTC.
		/// </summary>
		private void ServerAdmin_OnCustomCommand(IServerConnectionAdmin adm, string cmd, string prm)
		{
			Debug.WriteLine("Server sent a sys command '{0}' with param '{1}'", cmd, prm);
		}

		/// <summary>
		/// Обработчик изменений состояния коннекта
		/// </summary>
		/// <param name="admin"></param>
		private void ServerAdmin_OnStateChanged(ServerConnectionAdmin admin)
		{
			if (admin.LoggedIn)
				SetStatus(Status.LoggedIn);
			else if (admin.Connected)
				SetStatus(Status.Connected);
			else
				SetStatus(Status.Disconnected);

			if (admin.LoggedIn)
			{
				if (_currentRequest == null) return;
				// теперь, после логина, надо создать сессию, в рамках которой будет осуществляться подписка на онлайн-данные
				if (_requestSession == null)
				{
					_requestSession = admin.CreateSession();
					_requestSession.IsBulkSession = true; // это сессия, предназначеннная для подписки на онлайновые данные сразу по пачке инструментов
					_requestSession.OwnerSnapshotID = 0;  // необязательный идентификатор запроса
					_requestSession.OnConnect += Session_OnConnect;
					_requestSession.OnData += Session_OnData;
					_requestSession.OnDataBulk += Session_OnDataBulk;
					_requestSession.OnDisconnect += Session_OnDisconnect;
					_requestSession.OnError += Session_OnError;
					_requestSession.Connect();
				}
			}
		}

		/// <summary>
		/// Обработчик ошибок, при коннекте к серверу
		/// </summary>
		/// <param name="errors"></param>
		private void ServerAdmin_OnErrors(string[] errors)
		{
			SetStatus(Status.Error);
			foreach (string error in errors)
			{
				AppendText(error + "\r\n");
			}
		}

		/// <summary>
		/// Вызывается после открытия сессии.
		/// После этого можно отправлять запросы к серверу.
		/// </summary>
		private void Session_OnConnect(IServerSession session)
		{
			_currentStateReceived = false;
			_onlineBuffer = new ConcurrentQueue<ServerSessionData>();
			session.Params.Clear();

			// #version используется для идентификации версии сообщений, приходящих к клиенту через ServerSession.
			// Во время работы возможна переподписка с другими параметрами (например, другим набором колонок), 
			// поэтому надо отличать сообщения, приходящие по старой подписке, от приходящих по новой подписке.
			session.Params["#version"] = "1";
			session.Params["#return_subscription"] = "1";	// получить подписку (права) пользователя на инструменты
			session.Params["#return_columns"] = "1";		// получить обратно список колонок
			session.Params["#return_static_fields"] = "1";	// получить список статических полей

			if (!string.IsNullOrEmpty(_currentRequest.Columns))
			{
				// Ограничить список получаемых колонок, чтобы не гонять по сети лишние данные
				session.Params["#columns"] = _currentRequest.Columns;
			}
			if (!string.IsNullOrEmpty(_currentRequest.Instruments))
			{
				// Можно задать список инструментов через запятую. В противном случае будут приходить все инструменты указанной площадки.
				session.Params["#id_iss_array"] = _currentRequest.Instruments;
			}
			session.Query(string.Format("BULK7.{0}.Snapshot", (short)_currentRequest.Market));
		}

		/// <summary>
		/// Флаг, указывающий на то, что с сервера была получена первая порция данных (OnDataBulk).
		/// </summary>
		private bool _currentStateReceived;

		/// <summary>
		/// Вызывается один раз, после выполнения подписки при поступлении первой порции данных.
		/// Первая порция содержит в себе ещё не онлайновые данные, а результат выполнения команды подписки.
		/// </summary>
		private void Session_OnDataBulk(IServerSession session)
		{
			try
			{
				ServerSessionData data = session.Data;
				if (data.SelectSection(SectionParams))
				{
					for (int rowIndex = 0; rowIndex < data.Rows; rowIndex++)
					{
						string paramName = data[rowIndex, 0];
						string paramVal = data[rowIndex, 1];
						Debug.WriteLine("{0} = {1}", paramName, paramVal);
						if (paramName == "#columns")
						{
							// Если в запросе на сервер не отправлялся список колонок #columns,
							// то здесь мы получим все доступные колонки.
							_columns = paramVal.Split(',');
						}
						else if (paramName == "#result")
						{
							string[] resultParts = paramVal.Split('|');
							if (resultParts.Length > 0)
							{
								int resultCode = Int32.Parse(resultParts[0]);
								if (resultCode < 0 && resultParts.Length > 1)
								{
									string messageText = resultParts[1];
									AppendText(messageText + "\r\n");
								}
							}
						}
					}
				}
				ProcessSessionData(data);
				_currentStateReceived = true;
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex.ToString());
			}
		}

		/// <summary>
		/// Буфер онлайновых данных
		/// </summary>
		private ConcurrentQueue<ServerSessionData> _onlineBuffer;

		/// <summary>
		/// Вызывается при поступлении порции онлайновых данных
		/// </summary>
		private void Session_OnData(IServerSession session)
		{
			ServerSessionData data = session.Data;
			if (_onlineBuffer != null)
			{
				_onlineBuffer.Enqueue(data);
			}
		}

		/// <summary>
		/// Разбирает данные, полученные с сервера
		/// </summary>
		private void ProcessSessionData(ServerSessionData data)
		{
			if (data.SelectSection(SectionData))
			{
				StringBuilder sb = new StringBuilder();
				int nCount = 0;
				for (int rowIndex = 0; rowIndex < data.Rows; rowIndex++)
				{
					for (int colIndex = 0; colIndex < _columns.Length; colIndex++)
					{
						if (_columns[colIndex] == "-") continue;	// Пропускаем неизвестные колонки, запрошенные пользователем
						string colValue = data[rowIndex, colIndex];
						sb.AppendFormat("{0}:{1}, ", _columns[colIndex], colValue);
					}
					sb.AppendLine();
					nCount++;
					if (nCount < 100) continue;
					AppendText(sb.ToString());
					nCount = 0;
					sb.Clear();
				}
				if (nCount > 0)
				{
					AppendText(sb.ToString());
				}
			}
		}

		/// <summary>
		/// Сессия была закрыта
		/// </summary>
		/// <param name="session"></param>
		void Session_OnDisconnect(IServerSession session)
		{
			SetStatus(Status.Disconnected);
		}

		/// <summary>
		/// Ошибка в запросе, отправленном в сессию
		/// </summary>
		/// <param name="session"></param>
		void Session_OnError(IServerSession session)
		{
			SetStatus(Status.Error);
			AppendText(session.ErrorText + "\r\n");
		}

		private void btnStart_Click(object sender, EventArgs e)
		{
			if (_isRunning)
			{
				// При повторном нажатии на кнопку - останавливаем процесс
				_cancellation.Cancel();
				return;
			}
			try
			{
				SetButtonText("Stop");
				txtOutput.Text = "";
				_cancellation = new CancellationTokenSource();
				CancellationToken token = _cancellation.Token;
				_isRunning = true;
				_currentRequest = new Request();
				_currentRequest.Market =
					(MarketFeedGroup) Enum.Parse(typeof (MarketFeedGroup), (string) cbMarketFeedGroup.SelectedItem);
				if (chkOnlyColumns.Checked)
				{
					_currentRequest.Columns = txtColumns.Text;
				}
				if (chkOnlyInstruments.Checked)
				{
					_currentRequest.Instruments = txtInstruments.Text;
				}

				Task.Factory.StartNew(() =>
				{
					try
					{
						// Выполняем подключение к серверу, а дальше по цепочке обработчиков произойдет вызов методов:
						// - ServerAdmin_OnStateChanged
						// - Session_OnConnect
						// - Session_OnDataBulk
						// - Session_OnData (многократно)
						Connect();
						while (true)
						{
							token.ThrowIfCancellationRequested();
							Thread.Sleep(10);
							if (_currentStateReceived && _onlineBuffer != null)
							{
								ServerSessionData data;
								while (_onlineBuffer.TryDequeue(out data))
								{
									token.ThrowIfCancellationRequested();
									ProcessSessionData(data);
								}
							}
						}
					}
					catch (OperationCanceledException)
					{
						Debug.WriteLine("Task was cancelled");
					}
					catch (Exception ex)
					{
						MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
						SetStatus(Status.Error);
					}
					finally
					{
						Disconnect();
						SetButtonText("Start");
						_isRunning = false;
					}
				}
				, token);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		/// <summary>
		/// Логгирует изменения статуса коннекта к серверу и выводит его в окно
		/// </summary>
		/// <param name="status">Новое значение статуса</param>
		private void SetStatus(Status status)
		{
			if (_status == status) return;
			_status = status;
			string statusName = _status.ToString();
			Debug.WriteLine("Status: " + statusName);
			ShowStatus(statusName);
		}

		private delegate void ShowStatusCallback(string text);
		private void ShowStatus(string text)
		{
			if (lblStatus.InvokeRequired)
			{
				ShowStatusCallback d = ShowStatus;
				lblStatus.Invoke(d, new object[] { text });
			}
			else
			{
				lblStatus.Text = text;
			}
		}

		private delegate void SetButtonTextCallback(string text);
		private void SetButtonText(string text)
		{
			if (btnStart.InvokeRequired)
			{
				SetButtonTextCallback d = SetButtonText;
				btnStart.Invoke(d, new object[] { text });
			}
			else
			{
				btnStart.Text = text;
			}
		}

		private delegate void AppendTextCallback(string text);
		private void AppendText(string text)
		{
			if (txtOutput.InvokeRequired)
			{
				AppendTextCallback d = AppendText;
				txtOutput.Invoke(d, new object[] { text });
			}
			else
			{
				txtOutput.AppendText(text);
			}
		}
	}
}
