﻿using System;
using System.Diagnostics;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Interfax.Efir.MarketModel;
using Interfax.Efir.PartnerDataModel;
using Interfax.Efir.PartnerDataModel.DataContracts;
using PartnerDataClient.Properties;

namespace PartnerDataClient.Pages
{
	/// <summary>
	/// Sample form for IGuideService.GetInstrumentGuide method in EfirPartnerDataService.
	/// </summary>
	public partial class InstrumentGuidePage : Form
	{
		public InstrumentGuidePage()
		{
			InitializeComponent();
		}

		protected override void OnLoad(EventArgs e)
		{
			base.OnLoad(e);
			// Fill combo-box with possible values of MarketPlace
			foreach (var name in Enum.GetNames(typeof(MarketPlace)))
			{
				cbMarketplace.Items.Add(name);
			}
			cbMarketplace.SelectedItem = "MICEX";
		}

		protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
		{
			if (_isRunning)
			{
				Debug.WriteLine("Stopping InstrumentGuidePage");
				_cancellation.Cancel();
				while (_isRunning)
				{
					Application.DoEvents();
				}
				Debug.WriteLine("InstrumentGuidePage is stopped");
			}
			base.OnClosing(e);
		}

		private void btnGet_Click(object sender, EventArgs e)
		{
			if (_isRunning)
			{
				// Stop the process
				_cancellation.Cancel();
				return;
			}

			try
			{
				SetButtonText("Stop");
				txtOutputFields.Text = "";
				txtOutputInstruments.Text = "";
				SetStatus("Loading...");
				_cancellation = new CancellationTokenSource();
				CancellationToken token = _cancellation.Token;
				_isRunning = true;
				MarketPlace mp = (MarketPlace)Enum.Parse(typeof(MarketPlace), (string)cbMarketplace.SelectedItem);
				DateTime instrUpdatedDate = (dtInstrUpdateDate.Checked) ? dtInstrUpdateDate.Value.Date : DateTime.MinValue;
				DateTime fieldUpdatedDate = (dtFieldUpdateDate.Checked) ? dtFieldUpdateDate.Value.Date : DateTime.MinValue;

				Task.Factory.StartNew(() =>
										  {
											  try
											  {
												  // Make call to EfirPartnerDataService
												  IGuideService guideSrv = Program.ProxyManager.GetProxy<IGuideService>();
												  InstrumentGuide guide = guideSrv.GetInstrumentGuide(mp, instrUpdatedDate, fieldUpdatedDate);
												  token.ThrowIfCancellationRequested();
												  SetStatus(string.Format("Update date = {0}", guide.InstrumentsUpdatedDate));
												  if (guide.Fields != null && guide.Fields.Count > 0)
												  {
													  foreach (Field fld in guide.Fields)
													  {
														  token.ThrowIfCancellationRequested();
														  AppendField(fld + "\r\n");
													  }
												  }
												  else
												  {
													  AppendField("No fields\r\n");
												  }
												  if (guide.Instruments != null && guide.Instruments.Count > 0)
												  {
													  int nCount = 0;
													  StringBuilder sb = null;
													  foreach (SecurityDefinition secDef in guide.Instruments)
													  {
														  token.ThrowIfCancellationRequested();
														  sb = sb ?? new StringBuilder();
														  sb.AppendLine(secDef.ToString());
														  nCount++;
														  if (nCount < 100) continue;
														  AppendInstrument(sb.ToString());
														  sb = null;
														  nCount = 0;
													  }
													  if (nCount > 0 && sb != null)
													  {
														  AppendInstrument(sb.ToString());
													  }
													  Debug.WriteLine("Instruments are displayed");
												  }
												  else
												  {
													  AppendInstrument("No instruments\r\n");
												  }
											  }
											  catch (OperationCanceledException)
											  {
												  Debug.WriteLine("Task was cancelled");
												  SetStatus("Cancelled");
											  }
											  catch (Exception ex)
											  {
												  MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
											  }
											  finally
											  {
												  SetButtonText("Get");
												  _isRunning = false;
											  }
										  }
									  , token);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private bool _isRunning;
		private CancellationTokenSource _cancellation = new CancellationTokenSource();

		private delegate void SetButtonTextCallback(string text);
		private void SetButtonText(string text)
		{
			if (btnGet.InvokeRequired)
			{
				SetButtonTextCallback d = SetButtonText;
				btnGet.Invoke(d, new object[] { text });
			}
			else
			{
				btnGet.Text = text;
			}
		}

		private delegate void AppendTextCallback(string text);
		private void AppendField(string text)
		{
			if (txtOutputFields.InvokeRequired)
			{
				AppendTextCallback d = AppendField;
				txtOutputFields.Invoke(d, new object[] { text });
			}
			else
			{
				txtOutputFields.AppendText(text);
			}
		}
		private void AppendInstrument(string text)
		{
			if (txtOutputInstruments.InvokeRequired)
			{
				AppendTextCallback d = AppendInstrument;
				txtOutputInstruments.Invoke(d, new object[] { text });
			}
			else
			{
				txtOutputInstruments.AppendText(text);
			}
		}

		private delegate void SetStatusCallback(string text);
		private void SetStatus(string text)
		{
			if (lblStatus.InvokeRequired)
			{
				SetStatusCallback d = SetStatus;
				lblStatus.Invoke(d, new object[] {text});
			}
			else
			{
				lblStatus.Text = text;
			}
		}
	}
}
