﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Configuration;
using System.Diagnostics;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Efir.Client.Communication;
using Interfax.Efir.MarketModel;
using PartnerDataClient.Properties;

namespace PartnerDataClient.Pages
{
	/// <summary>
	/// Пример обычной подписки на онлайн-данные
	/// </summary>
	public partial class SimpleOnlinePage : Form
	{
		private const string ClentVersion = "1.0.0.0";	//"7.9.4.50576";

		// Названия секций в ответах сервера
		const string SectionResult = "RESULT";
		const string SectionStaticFields = "HIDE_ONLINE";	// Массив флагов, отмещающих статические поля.
		const string SectionParams = "PARAM";		// Параметры запроса
		const string SectionData = "DATA";			// Собственно данные по инструмента
		const string SectionGlobalData = "GDATA";	// Пристутствует только для стаканов: в ней лежит id_iss

		private ServerConnectionAdmin _connAdmin;	// Менеджер коннектов к серверу
		private IServerSession _requestSession;		// Сессия
		private Status _status = Status.Offline;	// Текущее состояние сессии
		private bool _isRunning;					// Флаг активности
		private CancellationTokenSource _cancellation = new CancellationTokenSource();
		private Request _currentRequest;			// Параметры текущего выполняемого запроса

		/// <summary>
		/// Состояние сессии
		/// </summary>
		private enum Status
		{
			Offline,
			Connecting,
			LoggedIn,
			Connected,
			Disconnected,
			Error
		}

		/// <summary>
		/// Параметры запроса, введённые пользователем
		/// </summary>
		private class Request
		{
			/// <summary>
			/// Торговая площадка
			/// </summary>
			public MarketFeedGroup Market { get; set; }
			/// <summary>
			/// Список колонок
			/// </summary>
			public string[] Columns { get; set; }
			/// <summary>
			/// Список инструментов
			/// </summary>
			public string[] Instruments { get; set; }
		}

		public SimpleOnlinePage()
		{
			InitializeComponent();
		}

		protected override void OnLoad(EventArgs e)
		{
			base.OnLoad(e);
			SetStatus(Status.Offline);
			// Заполняем выпадающий список значениями MarketFeedGroup
			List<string> groups = new List<string>(Enum.GetNames(typeof(MarketFeedGroup)));
			groups.Sort();
			foreach (var name in groups)
			{
				cbMarketFeedGroup.Items.Add(name);
			}
			cbMarketFeedGroup.SelectedItem = "MicexShares";
			// Тестовый набор инструментов с фида MicexShares
			txtInstruments.Text =
				"108217,35154,81449,93762,125341,32941,99209,115591,83967,83968,82730,32942,107106,107107,125342,125343,125344,125345,126069,126070,124059,32945,32946,32947,44927,44928,44929,128438,82271,82272,84794,84795,97089,143921,128009,128010,81026,128011,93496,93497,113107,84566,84567,125346,128012,85207,96319,96320,99087,81055";
		}

		protected override void OnClosing(CancelEventArgs e)
		{
			if (_isRunning)
			{
				// Если выполнялся запрос - посылаем сигнал и дожидаемся его остановки
				Debug.WriteLine("Stopping SimpleOnlinePage");
				_cancellation.Cancel();
				while (_isRunning)
				{
					Application.DoEvents();
				}
				Debug.WriteLine("SimpleOnlinePage is stopped");
			}
			base.OnClosing(e);
		}

		/// <summary>
		/// Подключение к online-серверу
		/// </summary>
		private void Connect()
		{
			try
			{
				SetStatus(Status.Connecting);
				_connAdmin = ServerConnectionAdmin.Create(ClentVersion, AsyncOperationManager.SynchronizationContext);
				_connAdmin.OnCustomCommand += ServerAdmin_OnCustomCommand;
				_connAdmin.OnErrors += ServerAdmin_OnErrors;
				_connAdmin.OnStateChanged += ServerAdmin_OnStateChanged;

				string onlineServerAddress = ConfigurationManager.AppSettings["OnlineServer"];
				ConnectionInfo connectionInfo = new ConnectionInfo(onlineServerAddress, Program.UserName, Program.Password, "NEWS");
				_connAdmin.ConnectFor(connectionInfo);
			}
			catch (Exception ex)
			{
				SetStatus(Status.Error);
				AppendText(ex.ToString());
			}
		}

		/// <summary>
		/// Отключение от сервера - закрываем сессию и коннект к серверу
		/// </summary>
		private void Disconnect()
		{
			if (_requestSession != null)
			{
				if (_requestSession.IsConnected)
				{
					_requestSession.Disconnect();
				}
				_requestSession.OnConnect -= Session_OnConnect;
				_requestSession.OnData -= Session_OnData;
				_requestSession.OnDisconnect -= Session_OnDisconnect;
				_requestSession.OnError -= Session_OnError;
				_requestSession.Dispose();
				_requestSession = null;
			}
			if (_connAdmin != null)
			{
				if (_connAdmin.Connected)
				{
					_connAdmin.Disconnect();
				}
				_connAdmin.OnCustomCommand -= ServerAdmin_OnCustomCommand;
				_connAdmin.OnErrors -= ServerAdmin_OnErrors;
				_connAdmin.OnStateChanged -= ServerAdmin_OnStateChanged;
				_connAdmin.Dispose();
				_connAdmin = null;
			}
			SetStatus(Status.Offline);
		}

		/// <summary>
		/// Обработчик системных команд от сервера.
		/// Чаще всего это бывает команда SYS_TIME, параметром которой является серверное время в UTC.
		/// </summary>
		private void ServerAdmin_OnCustomCommand(IServerConnectionAdmin adm, string cmd, string prm)
		{
			Debug.WriteLine("Server sent a sys command '{0}' with param '{1}'", cmd, prm);
		}

		/// <summary>
		/// Обработчик изменений состояния коннекта
		/// </summary>
		/// <param name="admin"></param>
		private void ServerAdmin_OnStateChanged(ServerConnectionAdmin admin)
		{
			if (admin.LoggedIn)
				SetStatus(Status.LoggedIn);
			else if (admin.Connected)
				SetStatus(Status.Connected);
			else
				SetStatus(Status.Disconnected);

			if (admin.LoggedIn)
			{
				if (_currentRequest == null) return;
				// теперь, после логина, надо создать сессию, в рамках которой будет осуществляться подписка на онлайн-данные
				if (_requestSession == null)
				{
					_requestSession = admin.CreateSession();
					_requestSession.IsBulkSession = false;
					_requestSession.OwnerSnapshotID = 0;  // необязательный идентификатор запроса
					_requestSession.OnConnect += Session_OnConnect;
					_requestSession.OnData += Session_OnData;
					_requestSession.OnDisconnect += Session_OnDisconnect;
					_requestSession.OnError += Session_OnError;
					_requestSession.Connect();
				}
			}
		}

		/// <summary>
		/// Обработчик ошибок, при коннекте к серверу
		/// </summary>
		/// <param name="errors"></param>
		private void ServerAdmin_OnErrors(string[] errors)
		{
			SetStatus(Status.Error);
			foreach (string error in errors)
			{
				AppendText(error + "\r\n");
			}
		}

		/// <summary>
		/// Вызывается после открытия сессии.
		/// После этого можно отправлять запросы к серверу.
		/// </summary>
		private void Session_OnConnect(IServerSession session)
		{
			session.Params.Clear();

			// #version используется для идентификации версии сообщений, приходящих к клиенту через ServerSession.
			// Во время работы возможна переподписка с другими параметрами (например, другим набором колонок), 
			// поэтому надо отличать сообщения, приходящие по старой подписке, от приходящих по новой подписке.
			session.Params["#version"] = "1";
			session.Params["#return_subscription"] = "1";	// получить подписку (права) пользователя на инструменты
			session.Params["#return_static_fields"] = "1";	// получить список статических полей
			session.Params["#columns"] = string.Join(",", _currentRequest.Columns);				// список загружаемых колонок
			if (_currentRequest.Instruments.Length == 1)
			{
				session.Params["#id_iss"] = _currentRequest.Instruments[0];
				session.Query(string.Format("QUOTE7.{0}.Snapshot", (short)_currentRequest.Market));
			}
			else
			{
				// список инструментов через запятую
				session.Params["#id_iss_array"] = string.Join(",", _currentRequest.Instruments);
				session.Query(string.Format("QUOTE7.{0}.SnapshotMulti", (short) _currentRequest.Market));
			}
		}

		/// <summary>
		/// Вызывается при поступлении порции онлайновых данных
		/// </summary>
		private void Session_OnData(IServerSession session)
		{
			ProcessSessionData(session.Data);
		}

		/// <summary>
		/// Разбирает данные, полученные с сервера
		/// </summary>
		private void ProcessSessionData(ServerSessionData data)
		{
			if (data.SelectSection(SectionData))
			{
				StringBuilder sb = new StringBuilder();
				int nCount = 0;
				for (int rowIndex = 0; rowIndex < data.Rows; rowIndex++)
				{
					for (int colIndex = 0; colIndex < _currentRequest.Columns.Length; colIndex++)
					{
						string colValue = data[rowIndex, colIndex];
						sb.AppendFormat("{0}:{1}, ", _currentRequest.Columns[colIndex], colValue);
					}
					sb.AppendLine();
					nCount++;
					if (nCount < 100) continue;
					AppendText(sb.ToString());
					nCount = 0;
					sb.Clear();
				}
				if (nCount > 0)
				{
					AppendText(sb.ToString());
				}
			}
		}

		/// <summary>
		/// Сессия была закрыта
		/// </summary>
		/// <param name="session"></param>
		void Session_OnDisconnect(IServerSession session)
		{
			SetStatus(Status.Disconnected);
		}

		/// <summary>
		/// Ошибка в запросе, отправленном в сессию
		/// </summary>
		/// <param name="session"></param>
		void Session_OnError(IServerSession session)
		{
			SetStatus(Status.Error);
			AppendText(session.ErrorText + "\r\n");
		}

		private void btnStart_Click(object sender, EventArgs e)
		{
			if (_isRunning)
			{
				// При повторном нажатии на кнопку - останавливаем процесс
				_cancellation.Cancel();
				return;
			}
			try
			{
				SetButtonText("Stop");
				txtOutput.Text = "";
				_cancellation = new CancellationTokenSource();
				CancellationToken token = _cancellation.Token;
				_isRunning = true;
				_currentRequest = new Request();
				_currentRequest.Market =
					(MarketFeedGroup) Enum.Parse(typeof (MarketFeedGroup), (string) cbMarketFeedGroup.SelectedItem);
				_currentRequest.Columns = txtColumns.Text.Split(',');
				_currentRequest.Instruments = txtInstruments.Text.Split(',');

				Task.Factory.StartNew(() =>
				{
					try
					{
						// Выполняем подключение к серверу, а дальше по цепочке обработчиков произойдет вызов методов:
						// - ServerAdmin_OnStateChanged
						// - Session_OnConnect
						// - Session_OnData (многократно)
						Connect();
						while (true)
						{
							token.ThrowIfCancellationRequested();
							Thread.Sleep(10);
						}
					}
					catch (OperationCanceledException)
					{
						Debug.WriteLine("Task was cancelled");
					}
					catch (Exception ex)
					{
						MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
						SetStatus(Status.Error);
					}
					finally
					{
						Disconnect();
						SetButtonText("Start");
						_isRunning = false;
					}
				}
				, token);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		/// <summary>
		/// Логгирует изменения статуса коннекта к серверу и выводит его в окно
		/// </summary>
		/// <param name="status">Новое значение статуса</param>
		private void SetStatus(Status status)
		{
			if (_status == status) return;
			_status = status;
			string statusName = _status.ToString();
			Debug.WriteLine("Status: " + statusName);
			ShowStatus(statusName);
		}

		private delegate void ShowStatusCallback(string text);
		private void ShowStatus(string text)
		{
			if (lblStatus.InvokeRequired)
			{
				ShowStatusCallback d = ShowStatus;
				lblStatus.Invoke(d, new object[] { text });
			}
			else
			{
				lblStatus.Text = text;
			}
		}

		private delegate void SetButtonTextCallback(string text);
		private void SetButtonText(string text)
		{
			if (btnStart.InvokeRequired)
			{
				SetButtonTextCallback d = SetButtonText;
				btnStart.Invoke(d, new object[] { text });
			}
			else
			{
				btnStart.Text = text;
			}
		}

		private delegate void AppendTextCallback(string text);
		private void AppendText(string text)
		{
			if (txtOutput.InvokeRequired)
			{
				AppendTextCallback d = AppendText;
				txtOutput.Invoke(d, new object[] { text });
			}
			else
			{
				txtOutput.AppendText(text);
			}
		}
	}
}
