﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Ifx.Helpers;
using Interfax.Efir.MarketModel;
using Interfax.Efir.PartnerDataModel;
using Interfax.Efir.PartnerDataModel.DataContracts;
using Interfax.Efir.PartnerDataModel.DataContracts.Micex;
using Interfax.Efir.PartnerDataModel.DataContracts.Rts;
using Interfax.Efir.PartnerDataModel.DataContracts.SP;
using PartnerDataClient.Properties;

namespace PartnerDataClient.Pages
{
	public partial class SnapshotsPage : Form
	{
		public SnapshotsPage()
		{
			InitializeComponent();
		}
		protected override void OnLoad(EventArgs e)
		{
			base.OnLoad(e);
			// Fill combo-box with supported values of MarketFeedGroup
			cbMarketFeedGroup.Items.Add("MicexCBonds");
			cbMarketFeedGroup.Items.Add("MicexCurrency");
			cbMarketFeedGroup.Items.Add("MicexIndex");
			cbMarketFeedGroup.Items.Add("MicexShares");
			cbMarketFeedGroup.Items.Add("RtsCBonds");
			cbMarketFeedGroup.Items.Add("RtsFutures");
			cbMarketFeedGroup.Items.Add("RtsIndex");
			cbMarketFeedGroup.Items.Add("RtsShares");
			cbMarketFeedGroup.Items.Add("Forex");
			cbMarketFeedGroup.SelectedIndex = 0;
		}

		protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
		{
			if (_isRunning)
			{
				Debug.WriteLine("Stopping SnapshotsPage");
				_cancellation.Cancel();
				while (_isRunning)
				{
					Application.DoEvents();
				}
				Debug.WriteLine("SnapshotsPage is stopped");
			}
			base.OnClosing(e);
		}

		private void btnGet_Click(object sender, EventArgs e)
		{
			if (_isRunning)
			{
				// Stop the process
				_cancellation.Cancel();
				return;
			}

			try
			{
				SetButtonText("Stop");
				txtOutput.Text = "";
				SetStatus("Loading...");
				_cancellation = new CancellationTokenSource();
				CancellationToken token = _cancellation.Token;
				_isRunning = true;
				MarketFeedGroup marketFeedGroup = (MarketFeedGroup)Enum.Parse(typeof(MarketFeedGroup), (string)cbMarketFeedGroup.SelectedItem);

				Task.Factory.StartNew(() =>
					{
					Stream snapshotStm = null;
					try
					{
						// Make call to EfirPartnerDataService
						IHistoryRecoveryService histSrv = Program.ProxyManager.GetProxy<IHistoryRecoveryService>();
						snapshotStm = histSrv.RecieveSnapshots(marketFeedGroup);
						switch (marketFeedGroup)
						{
							case MarketFeedGroup.MicexShares:
								ReadSnapshots<MicexSharesSnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.MicexCBonds:
								ReadSnapshots<MicexCorpBondsSnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.MicexCurrency:
								ReadSnapshots<MicexCurrencySnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.MicexIndex:
								ReadSnapshots<MicexIndexSnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.RtsShares:
								ReadSnapshots<RtsSharesSnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.RtsCBonds:
								ReadSnapshots<RtsCorpBondsSnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.RtsFutures:
								ReadSnapshots<RtsFuturesSnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.RtsIndex:
								ReadSnapshots<RtsIndexSnapshot>(snapshotStm, token);
								break;
							case MarketFeedGroup.Forex:
								ReadSnapshots<SpForexSnapshot>(snapshotStm, token);
								break;
							default:
								SetStatus("Loaded");
								// Unsupported type of snapshots 
								break;
						}
					}
					catch (OperationCanceledException)
					{
						Debug.WriteLine("Task was cancelled");
						SetStatus("Cancelled");
					}
					catch (Exception ex)
					{
						MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
						SetStatus("Error");
					}
					finally
					{
						if (snapshotStm != null)
						{
							try
							{
								snapshotStm.Dispose();
							}
							catch (TimeoutException)
							{
								// В биндинге выставлен closeTimeout="0:00:01", чтобы TimeoutException прилетел быстрее, когда мы закрываем stream не дожидаясь окончания данных.
								Debug.WriteLine("The stream is cancelled");
							}
						}
						SetButtonText("Get");
						_isRunning = false;
					}
				}
				, token);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private void ReadSnapshots<T>(Stream snapshotStm, CancellationToken token) where T : StreamedDtoBase
		{
			int totalCount = 0;
			var reader = new SerializedStreamReader<T>(snapshotStm);
			T snapshot;
			int nCount = 0;
			StringBuilder sb = null;
			while ((snapshot = reader.Next()) != null)
			{
				token.ThrowIfCancellationRequested();
				sb = sb ?? new StringBuilder();
				sb.AppendLine(snapshot.ToString());
				nCount++;
				totalCount++;
				if (nCount < 100) continue;
				AppendText(sb.ToString());
				sb = null;
				nCount = 0;
				SetStatus(string.Format("Loaded {0}...", totalCount));
			}
			if (nCount > 0 && sb != null)
			{
				AppendText(sb.ToString());
			}
			SetStatus(string.Format("Loaded snapshots: {0}", totalCount));
		}

		private bool _isRunning;
		private CancellationTokenSource _cancellation = new CancellationTokenSource();

		private delegate void SetButtonTextCallback(string text);
		private void SetButtonText(string text)
		{
			if (btnGet.InvokeRequired)
			{
				SetButtonTextCallback d = SetButtonText;
				btnGet.Invoke(d, new object[] { text });
			}
			else
			{
				btnGet.Text = text;
			}
		}

		private delegate void AppendTextCallback(string text);
		private void AppendText(string text)
		{
			if (txtOutput.InvokeRequired)
			{
				AppendTextCallback d = AppendText;
				txtOutput.Invoke(d, new object[] { text });
			}
			else
			{
				txtOutput.AppendText(text);
			}
		}

		private delegate void SetStatusCallback(string text);
		private void SetStatus(string text)
		{
			if (lblStatus.InvokeRequired)
			{
				SetStatusCallback d = SetStatus;
				lblStatus.Invoke(d, new object[] { text });
			}
			else
			{
				lblStatus.Text = text;
			}
		}
	}
}
