﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Ifx.Helpers;
using Interfax.Efir.MarketModel;
using Interfax.Efir.PartnerDataModel;
using Interfax.Efir.PartnerDataModel.DataContracts;
using PartnerDataClient.Properties;
using System.IO;

namespace PartnerDataClient.Pages
{
	/// <summary>
	/// Пример загрузки истории сделок с использованием EfirPartnerDataService
	/// </summary>
	public partial class TradesHistoryPage : Form
	{
		public TradesHistoryPage()
		{
			InitializeComponent();
		}

		protected override void OnLoad(EventArgs e)
		{
			base.OnLoad(e);
			// Заполняем выпадающий список значениями MarketFeedGroup
			List<string> groups = new List<string>(Enum.GetNames(typeof(MarketFeedGroup)));
			groups.Sort();
			foreach (var name in groups)
			{
				cbMarketFeedGroup.Items.Add(name);
			}
			cbMarketFeedGroup.SelectedItem = "MicexShares";
		}

		protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
		{
			if (_isRunning)
			{
				// Если выполнялся запрос - посылаем сигнал и дожидаемся его остановки
				Debug.WriteLine("Stopping TradesHistoryPage");
				_cancellation.Cancel();
				while (_isRunning)
				{
					Application.DoEvents();
				}
				Debug.WriteLine("TradesHistoryPage is stopped");
			}
			base.OnClosing(e);
		}

		private void btnGet_Click(object sender, EventArgs e)
		{
			if (_isRunning)
			{
				// При повторном нажатии на кнопку - останавливаем процесс
				_cancellation.Cancel();
				return;
			}

			try
			{
				SetButtonText("Stop");
				txtOutput.Text = "";
				SetStatus("Loading...");
				_cancellation = new CancellationTokenSource();
				CancellationToken token = _cancellation.Token;
				_isRunning = true;
				MarketFeedGroup marketFeedGroup = (MarketFeedGroup)Enum.Parse(typeof(MarketFeedGroup), (string)cbMarketFeedGroup.SelectedItem);
				DateTime date = (dtDate.Checked) ? dtDate.Value.Date : DateTime.Today;
				int idIss;
				bool filterByInstrument = Int32.TryParse(txtIdIss.Text, out idIss);

				Task.Factory.StartNew(() =>
				{
					Stream tradesStm = null;
					try
					{
						// Выполняем вызов метода EfirPartnerDataService
						IHistoryRecoveryService histSrv = Program.ProxyManager.GetProxy<IHistoryRecoveryService>();
						tradesStm = histSrv.RecieveTrades(marketFeedGroup, date, date.AddDays(1));
						Trade trade;
						int totalCount = 0;
						int nCount = 0;
						StringBuilder sb = null;
						SerializedStreamReader<Trade> tradeReader = new SerializedStreamReader<Trade>(tradesStm);
						while ((trade = tradeReader.Next()) != null)
						{
							token.ThrowIfCancellationRequested();
							if (filterByInstrument)
							{
								if (trade.IssID != idIss) continue;
								AppendText(trade + "\r\n");
								totalCount++;
								SetStatus(string.Format("Loaded {0}...", totalCount));
								continue;
							}
							sb = sb ?? new StringBuilder();
							sb.AppendLine(trade.ToString());
							totalCount++;
							nCount++;
							if (nCount < 100) continue;
							AppendText(sb.ToString());
							sb = null;
							nCount = 0;
							SetStatus(string.Format("Loaded {0}...", totalCount));
						}
						if (nCount > 0 && sb != null)
						{
							AppendText(sb.ToString());
						}
						SetStatus(string.Format("Loaded trades: {0}", totalCount));
					}
					catch (OperationCanceledException)
					{
						Debug.WriteLine("Task was cancelled");
						SetStatus("Cancelled");
					}
					catch (Exception ex)
					{
						MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
						SetStatus("Error");
					}
					finally
					{
						if (tradesStm != null)
						{
							try
							{
								tradesStm.Dispose();
							}
							catch (TimeoutException)
							{
								// В биндинге выставлен closeTimeout="0:00:01", чтобы TimeoutException прилетел быстрее, когда мы закрываем stream не дожидаясь окончания данных.
								Debug.WriteLine("The stream is cancelled");
							}
						}
						SetButtonText("Get");
						_isRunning = false;
					}
				}
				, token);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, Resources.ErrorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private bool _isRunning;
		private CancellationTokenSource _cancellation = new CancellationTokenSource();

		private delegate void SetButtonTextCallback(string text);
		private void SetButtonText(string text)
		{
			if (btnGet.InvokeRequired)
			{
				SetButtonTextCallback d = SetButtonText;
				btnGet.Invoke(d, new object[] { text });
			}
			else
			{
				btnGet.Text = text;
			}
		}

		private delegate void AppendTextCallback(string text);
		private void AppendText(string text)
		{
			if (txtOutput.InvokeRequired)
			{
				AppendTextCallback d = AppendText;
				txtOutput.Invoke(d, new object[] { text });
			}
			else
			{
				txtOutput.AppendText(text);
			}
		}

		private delegate void SetStatusCallback(string text);
		private void SetStatus(string text)
		{
			if (lblStatus.InvokeRequired)
			{
				SetStatusCallback d = SetStatus;
				lblStatus.Invoke(d, new object[] { text });
			}
			else
			{
				lblStatus.Text = text;
			}
		}
	}
}
